;;;; lightning..scm
;
; Copyright (c) 2000-2005, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Unter den Gleichen 1
; 37130 Gleichen
; Germany


(eval-when (compile)
  (use srfi-13 srfi-69)
  (set! ##jit#jit-macros (make-hash-table eq?)) )

(cond-expand [hygienic-macros (include "chicken-match-macros")] [else])

(declare (hide jit:set-ip! jit:ip jit:flush jit:bytes))

(use easyffi)

#>
#include <lightning.h>
<#

(define-macro (jit:compile args . instrs)
  (define (dotype type)
    (case type
      [(integer) "int"]
      [(unsigned-integer unsigned-int) "int"]
      [(unsigned-char) "unsigned char"]
      [(unsigned-short) "unsigned short"]
      [(unsigned-long) "unsigned long"]
      [(insn jit:insn) "jit_insn *"]
      [else type] ) )
  (let ([locals '()]
	[argvars (map (match-lambda
			[(type var val) (cons var type)]
			[arg (syntax-error 'jit:compile "invalid argument" arg)] ) 
		      args) ] )
    (define (defined? v)
      (or (assq v argvars) (assq v locals)) )
    (define (extract-arg arg)
      (->string
       (cond [(number? arg) arg]
	     [(char? arg) (char->integer arg)]
	     [(string? arg) arg]
	     [(eq? 'forward arg) "jit_forward()"]
	     [(memq arg '(R0 R1 R2 V0 V1 V2 RET SP FPR0 FPR1 FPR2 FPR3 FPR4 FPR5)) (conc "JIT_" arg)]
	     [(defined? arg) arg]
	     [(symbol? arg) (syntax-error 'jit:compile "undefined variable" arg)]
	     [else (syntax-error 'jit:compile "invalid argument" arg)] ) ) )
    (define (localvar v)
      (sprintf "~a ~a;~%" (cdr v) (car v)) )
    (define single
      (match-lambda
	[((inst ...)) (single inst)]
	[(op args ...)
	 (let ([mx (hash-table-ref/default ##jit#jit-macros op #f)])
	   (if mx
	       (instruction (apply mx args))
	       (sprintf "jit_~a(~a);~%" op (string-intersperse (map extract-arg args) ", ")) ) ) ]
	[inst (syntax-error 'jit:compile "invalid instruction syntax" inst)] ) )
    (define instruction
      (match-lambda
	[('delay inst1 inst2)
	 (conc "jit_delay(" (instruction inst1) ", " (instruction inst2)) ]
	[('begin insts ...)
	 (string-intersperse (map instruction insts) "\n") ]
	[(type var '= . inst)
	 (unless (defined? var) 
	   (set! locals (alist-cons var (dotype type) locals)) )
	 (conc var " = " (single inst)) ]
	[(var '= . inst)
	 (instruction `(int ,var = ,@inst)) ]
	[(inst ...)
	 (single inst) ]
	[insn (syntax-error 'jit:compile "invalid instruction syntax" insn)] ) )
    (let ([insns (map instruction instrs)])
      `((foreign-lambda* void ,(map (lambda (a) (list (cdr a) (car a))) argvars)
	  ,(string-concatenate (map localvar locals))
	  ,(string-concatenate insns) )
	,@(map caddr args) ) ) ) )

(define-macro (define-jit-macro head . body)
  (match head
    [((? symbol? name) . llist)
     (let ([expander (eval `(lambda ,llist ,@body))])
       (hash-table-set! ##jit#jit-macros name expander)
       '(void) ) ]
    [_ (syntax-error 'define-jit-macro "invalid macro definition" `(define-jit-macro ,head ...))] ) )

(foreign-parse/declare #<<EOF
___declare(substitute, "C_jit_;jit:")

static void *C_jit_make_buffer(int size)
{
  return (void *)C_malloc(size * sizeof(jit_insn));
}

static void C_jit_flush_code_with_length(void *ptr, int len)
{
  jit_flush_code(ptr, (unsigned char *)ptr + len);
}

static unsigned int C_jit_byte_count(void *ptr1, void *ptr2)
{
  return (unsigned char *)ptr2 - (unsigned char *)ptr1;
}
EOF
)

(foreign-parse #<<EOF
void jit_set_ip(void *);
void jit_flush_code(void *, void *);
EOF
)

(declare (hide jit:make_buffer jit_flush_code jit_set_ip jit:byte_count))

(define (jit:make-buffer size)
  (or (jit:make_buffer size)
      (error "out of memory - can not allocate buffer") ) )

(define jit:release (foreign-lambda void "C_free" c-pointer))

(define-foreign-type jit:insn (pointer "jit_insn"))

(define (jit:ip)
  (foreign-value "jit_get_ip().ptr" c-pointer) )

(define (jit:flush start #!optional (end (jit:ip)))
  (if (fixnum? end)
      (jit:flush_code_with_length start end)
      (jit_flush_code start end) )
  start)

(define (jit:bytes start #!optional (end (jit:ip)))
  (jit:byte_count start end) )

(define (jit:set-ip! ptr)
  (jit_set_ip ptr) )
